from __future__ import absolute_import
from __future__ import print_function

import veriloggen.core.vtypes as vtypes
import veriloggen.core.module as module
import veriloggen.types.bram as bram
from veriloggen.seq.seq import Seq


def CoramMemoryInterface(m, name=None, datawidth=32, addrwidth=10,
                         p_addr='ADDR', p_rdata='Q', p_wdata='D', p_wenable='WE',
                         index=None):

    return bram.BramInterface(
        m, name=name, datawidth=datawidth, addrwidth=addrwidth,
        p_addr=p_addr, p_rdata=p_rdata, p_wdata=p_wdata, p_wenable=p_wenable, index=index)


def CoramMemorySlaveInterface(m, name=None, datawidth=32, addrwidth=10,
                              p_addr='ADDR', p_rdata='Q', p_wdata='D', p_wenable='WE',
                              index=None):

    return bram.BramSlaveInterface(
        m, name=name, datawidth=datawidth, addrwidth=addrwidth,
        p_addr=p_addr, p_rdata=p_rdata, p_wdata=p_wdata, p_wenable=p_wenable, index=index)


def CoramMemoryMasterInterface(m, name=None, datawidth=32, addrwidth=10,
                               p_addr='ADDR', p_rdata='Q', p_wdata='D', p_wenable='WE',
                               index=None):

    return bram.BramMasterInterface(
        m, name=name, datawidth=datawidth, addrwidth=addrwidth,
        p_addr=p_addr, p_rdata=p_rdata, p_wdata=p_wdata, p_wenable=p_wenable, index=index)


def mkCoramMemoryDefinition(numports):
    name = 'CoramMemory%dP' % numports
    m = module.Module(name)

    coram_thread_name = m.Parameter('CORAM_THREAD_NAME', 'none')
    coram_thread_id = m.Parameter('CORAM_THREAD_ID', 0)
    coram_id = m.Parameter('CORAM_ID', 0)
    coram_sub_id = m.Parameter('CORAM_SUB_ID', 0)
    coram_addr_len = m.Parameter('CORAM_ADDR_LEN', 10)
    coram_data_width = m.Parameter('CORAM_DATA_WIDTH', 32)

    clk = m.Input('CLK')

    interfaces = []

    for i in range(numports):
        index = i if numports > 1 else None
        interface = CoramMemorySlaveInterface(m, datawidth=coram_data_width,
                                              addrwidth=coram_addr_len, index=index)
        delay_addr_name = 'D_ADDR%d' % i if numports > 1 else 'D_ADDR'
        interface.delay_addr = m.Reg(delay_addr_name, coram_addr_len)
        interfaces.append(interface)

    mem = m.Reg('mem', coram_data_width, length=vtypes.Int(2)**coram_addr_len)

    for interface in interfaces:
        m.Always(vtypes.Posedge(clk))(
            vtypes.If(interface.wenable)(
                mem[interface.addr](interface.wdata)
            ),
            interface.delay_addr(interface.addr)
        )
        m.Assign(interface.rdata(mem[interface.delay_addr]))

    return m


coram_memroy_definitions = {}


def get_coram_memory_definition(numports):
    if numports in coram_memroy_definitions:
        return coram_memroy_definitions[numports]
    coram_memroy_definitions[numports] = mkCoramMemoryDefinition(numports)
    return coram_memroy_definitions[numports]


class CoramMemory(bram.Bram):

    def __init__(self, m, clk, rst,
                 thread_name, thread_id, id, sub_id=None,
                 datawidth=32, addrwidth=10, numports=1):

        self.m = m
        self.clk = clk
        self.rst = rst
        self.thread_name = thread_name
        self.thread_id = thread_id
        self.id = id
        self.sub_id = sub_id
        self.name = '_'.join([thread_name, 'memory', str(id)])
        self.datawidth = datawidth
        self.addrwidth = addrwidth

        self.interfaces = []
        for i in range(numports):
            index = i if numports > 1 else None
            obj = CoramMemoryInterface(m, name=self.name,
                                       datawidth=datawidth, addrwidth=addrwidth, index=index)
            self.interfaces.append(obj)

        self.definition = get_coram_memory_definition(numports)

        params = []
        params.append(('CORAM_THREAD_NAME', self.thread_name))
        params.append(('CORAM_THREAD_ID', self.thread_id))
        params.append(('CORAM_ID', self.id))
        if self.sub_id is not None:
            params.append(('CORAM_SUB_ID', self.sub_id))
        params.append(('CORAM_ADDR_LEN', self.addrwidth))
        params.append(('CORAM_DATA_WIDTH', self.datawidth))

        ports = []
        ports.append(('CLK', self.clk))
        ports.extend(m.connect_ports(self.definition, prefix=self.name + '_'))

        self.inst = self.m.Instance(
            self.definition, 'inst_' + self.name, params, ports)

        self.seq = Seq(m, self.name, clk, rst)

        self._write_disabled = [False for i in range(numports)]



from __future__ import absolute_import
from __future__ import print_function

import math
import veriloggen.core.vtypes as vtypes
from veriloggen.core.module import Module


def make_port(m, _type, *args, **kwargs):
    if 'initval' in kwargs and 'Reg' not in _type:
        del kwargs['initval']
    return getattr(m, _type)(*args, **kwargs)


class RAMInterface(object):
    _I = 'Input'
    _O = 'Output'

    def __init__(self, m, name=None, datawidth=32, addrwidth=10, itype=None, otype=None,
                 p_addr='addr', p_rdata='rdata', p_wdata='wdata', p_wenable='wenable',
                 index=None):

        if itype is None:
            itype = self._I
        if otype is None:
            otype = self._O

        self.m = m

        name_addr = p_addr if name is None else '_'.join([name, p_addr])
        name_rdata = p_rdata if name is None else '_'.join([name, p_rdata])
        name_wdata = p_wdata if name is None else '_'.join([name, p_wdata])
        name_wenable = p_wenable if name is None else '_'.join(
            [name, p_wenable])

        if index is not None:
            name_addr = name_addr + str(index)
            name_rdata = name_rdata + str(index)
            name_wdata = name_wdata + str(index)
            name_wenable = name_wenable + str(index)

        self.addr = make_port(m, itype, name_addr, addrwidth, initval=0)
        self.rdata = make_port(m, otype, name_rdata, datawidth, initval=0)
        self.wdata = make_port(m, itype, name_wdata, datawidth, initval=0)
        self.wenable = make_port(m, itype, name_wenable, initval=0)


def mkRAMCore(name, datawidth=32, addrwidth=10, numports=2):
    m = Module(name)
    clk = m.Input('CLK')

    interfaces = []

    for i in range(numports):
        interface = RAMInterface(
            m, name + '_%d' % i, datawidth, addrwidth)
        interface.delay_addr = m.Reg(name + '_%d_daddr' % i, addrwidth)
        interfaces.append(interface)

    mem = m.Reg('mem', datawidth, length=2**addrwidth)

    for interface in interfaces:
        m.Always(vtypes.Posedge(clk))(
            vtypes.If(interface.wenable)(
                mem[interface.addr](interface.wdata)
            ),
            interface.delay_addr(interface.addr)
        )
        m.Assign(interface.rdata(mem[interface.delay_addr]))

    return m


def mkRAM(name, datawidth=32, addrwidth=10, numports=2):
    if numports < 1:
        raise ValueError("numports must be greater than 0.")

    name = 'dataflow_ram_%d' % index
    m = mkRAMCore(name, datawidth, addrwidth, numports)

    return m


# global multiplier count
index_count = 0


def get_RAM(datawidth=32, length=1024, numports=1):
    global index_count
    ram = mkRAM(index_count, datawidth, length, numports)
    index_count += 1
    return ram


def reset():
    global index_count
    index_count = 0



from __future__ import absolute_import
from __future__ import print_function

from .memory import CoramMemory
from .memory import CoramMemoryInterface, CoramMemorySlaveInterface, CoramMemoryMasterInterface

from .channel import CoramChannel
from .channel import CoramChannelWriteInterface, CoramChannelWriteSlaveInterface, CoramChannelWriteMasterInterface
from .channel import CoramChannelReadInterface, CoramChannelReadSlaveInterface, CoramChannelReadMasterInterface



from __future__ import absolute_import
from __future__ import print_function
import veriloggen
import types_coram_memory

expected_verilog = """
module test;

  reg CLK;
  reg RST;

  main
  uut
  (
    .CLK(CLK),
    .RST(RST)
  );


  initial begin
    $dumpfile("uut.vcd");
    $dumpvars(0, uut, CLK, RST);
  end


  initial begin
    CLK = 0;
    forever begin
      #5 CLK = !CLK;
    end
  end


  initial begin
    RST = 0;
    #100;
    RST = 1;
    #100;
    RST = 0;
    #100000;
    $finish;
  end


endmodule



module main
(
  input CLK,
  input RST
);

  reg [14-1:0] cthread_memory_0_ADDR;
  wire [32-1:0] cthread_memory_0_Q;
  reg [32-1:0] cthread_memory_0_D;
  reg cthread_memory_0_WE;

  CoramMemory1P
  #(
    .CORAM_THREAD_NAME("cthread"),
    .CORAM_THREAD_ID(0),
    .CORAM_ID(0),
    .CORAM_ADDR_LEN(14),
    .CORAM_DATA_WIDTH(32)
  )
  inst_cthread_memory_0
  (
    .CLK(CLK),
    .ADDR(cthread_memory_0_ADDR),
    .Q(cthread_memory_0_Q),
    .D(cthread_memory_0_D),
    .WE(cthread_memory_0_WE)
  );

  reg [32-1:0] count;
  reg [32-1:0] sum;
  reg [32-1:0] addr;
  reg [32-1:0] fsm;
  localparam fsm_init = 0;
  reg _cthread_memory_0_cond_0_1;
  reg _tmp_0;
  reg _cthread_memory_0_cond_1_1;
  reg _cthread_memory_0_cond_2_1;
  reg _cthread_memory_0_cond_2_2;
  reg [32-1:0] _d1_fsm;
  reg _fsm_cond_2_0_1;
  reg _fsm_cond_3_1_1;
  localparam fsm_1 = 1;
  localparam fsm_2 = 2;
  localparam fsm_3 = 3;

  always @(posedge CLK) begin
    if(RST) begin
      fsm <= fsm_init;
      _d1_fsm <= fsm_init;
      addr <= 0;
      count <= 0;
      sum <= 0;
      _fsm_cond_2_0_1 <= 0;
      _fsm_cond_3_1_1 <= 0;
    end else begin
      _d1_fsm <= fsm;
      case(_d1_fsm)
        fsm_2: begin
          if(_fsm_cond_2_0_1) begin
            $display("sum=%d", sum);
          end 
        end
        fsm_3: begin
          if(_fsm_cond_3_1_1) begin
            $display("sum=%d", sum);
          end 
        end
      endcase
      case(fsm)
        fsm_init: begin
          addr <= 0;
          count <= 0;
          sum <= 0;
          fsm <= fsm_1;
        end
        fsm_1: begin
          addr <= addr + 1;
          count <= count + 1;
          if(count == 15) begin
            addr <= 0;
            count <= 0;
          end 
          if(count == 15) begin
            fsm <= fsm_2;
          end 
        end
        fsm_2: begin
          addr <= addr + 1;
          count <= count + 1;
          if(_tmp_0) begin
            sum <= sum + cthread_memory_0_Q;
          end 
          _fsm_cond_2_0_1 <= _tmp_0;
          if(count == 15) begin
            addr <= 0;
            count <= 0;
          end 
          if(count == 15) begin
            fsm <= fsm_3;
          end 
        end
        fsm_3: begin
          if(_tmp_0) begin
            sum <= sum + cthread_memory_0_Q;
          end 
          _fsm_cond_3_1_1 <= _tmp_0;
        end
      endcase
    end
  end


  always @(posedge CLK) begin
    if(RST) begin
      cthread_memory_0_ADDR <= 0;
      cthread_memory_0_D <= 0;
      cthread_memory_0_WE <= 0;
      _cthread_memory_0_cond_0_1 <= 0;
      _cthread_memory_0_cond_1_1 <= 0;
      _tmp_0 <= 0;
      _cthread_memory_0_cond_2_1 <= 0;
      _cthread_memory_0_cond_2_2 <= 0;
    end else begin
      if(_cthread_memory_0_cond_2_2) begin
        _tmp_0 <= 0;
      end 
      if(_cthread_memory_0_cond_0_1) begin
        cthread_memory_0_WE <= 0;
      end 
      if(_cthread_memory_0_cond_1_1) begin
        _tmp_0 <= 1;
      end 
      _cthread_memory_0_cond_2_2 <= _cthread_memory_0_cond_2_1;
      if(fsm == 1) begin
        cthread_memory_0_ADDR <= addr;
        cthread_memory_0_D <= count;
        cthread_memory_0_WE <= 1;
      end 
      _cthread_memory_0_cond_0_1 <= fsm == 1;
      if(fsm == 2) begin
        cthread_memory_0_ADDR <= addr;
      end 
      _cthread_memory_0_cond_1_1 <= fsm == 2;
      _cthread_memory_0_cond_2_1 <= fsm == 2;
    end
  end


endmodule



module CoramMemory1P #
(
  parameter CORAM_THREAD_NAME = "none",
  parameter CORAM_THREAD_ID = 0,
  parameter CORAM_ID = 0,
  parameter CORAM_SUB_ID = 0,
  parameter CORAM_ADDR_LEN = 10,
  parameter CORAM_DATA_WIDTH = 32
)
(
  input CLK,
  input [CORAM_ADDR_LEN-1:0] ADDR,
  output [CORAM_DATA_WIDTH-1:0] Q,
  input [CORAM_DATA_WIDTH-1:0] D,
  input WE
);

  reg [CORAM_ADDR_LEN-1:0] D_ADDR;
  reg [CORAM_DATA_WIDTH-1:0] mem [0:2**CORAM_ADDR_LEN-1];

  always @(posedge CLK) begin
    if(WE) begin
      mem[ADDR] <= D;
    end 
    D_ADDR <= ADDR;
  end

  assign Q = mem[D_ADDR];

endmodule
"""


def test():
    veriloggen.reset()
    test_module = types_coram_memory.mkTest()
    code = test_module.to_verilog()

    from pyverilog.vparser.parser import VerilogParser
    from pyverilog.ast_code_generator.codegen import ASTCodeGenerator
    parser = VerilogParser()
    expected_ast = parser.parse(expected_verilog)
    codegen = ASTCodeGenerator()
    expected_code = codegen.visit(expected_ast)

    assert(expected_code == code)



#-------------------------------------------------------------------------------
# Veriloggen: A library for constructing a Verilog HDL source code in Python
# 
# Copyright (C) 2015, Shinya Takamaeda-Yamazaki
# License: Apache 2.0
#-------------------------------------------------------------------------------
from __future__ import absolute_import
from __future__ import print_function

# Verilog HDL Core
from .core.vtypes import *
from .core.module import Module, StubModule, Instance, GenerateFor, GenerateIf, connect_same_name
from .core.function import Function, FunctionCall
from .core.task import Task, TaskCall

# Verilog 
from .verilog import from_verilog
from .simulation import simulation

# Extension
from .seq.seq import Seq, TmpSeq, make_condition
from .fsm.fsm import FSM, TmpFSM
from .pipeline.pipeline import Pipeline

# Extension reset
from .seq.seq import reset as seq_reset
from .fsm.fsm import reset as fsm_reset
from .dataflow.dataflow import reset as dataflow_reset

def reset():
    seq_reset()
    fsm_reset()
    dataflow_reset()



from __future__ import absolute_import
from __future__ import print_function

import os
import sys

import veriloggen.core.vtypes as vtypes

class ResetVisitor(object):
    def generic_visit(self, node):
        raise TypeError("Type %s is not supported." % str(type(node)))
    
    def visit(self, node):
        if isinstance(node, vtypes._Variable):
            return self.visit__Variable(node)
        
        visitor = getattr(self, 'visit_' + node.__class__.__name__, self.generic_visit)
        return visitor(node)

    def visit__Variable(self, node):
        if node.initval is None:
            return None
        return vtypes.Subst(node, node.initval)

    def visit_Pointer(self, node):
        val = self.visit(node.var)
        if val is None:
            return None
        return vtypes.Subst(vtypes.Pointer(val.left, node.pos), 
                            vtypes.Pointer(val.right, node.pos))

    def visit_Slice(self, node):
        val = self.visit(node.var)
        if val is None:
            return None
        return vtypes.Subst(vtypes.Slice(val.left, node.msb, node.lsb),
                            vtypes.Slice(val.right, node.msb, node.lsb))

    def visit_Cat(self, node):
        left = []
        right = []
        for v in node.vars:
            val = self.visit(v)
            right = vtypes.IntX() if val is None else val.right
            left.append(v)
            right.append(right)
        return vtypes.Subst(vtypes.Cat(tuple(left)), vtypes.Cat(tuple(right)))



from __future__ import absolute_import
from __future__ import print_function

import veriloggen.core.vtypes as vtypes
import veriloggen.core.module as module

def mkMultiplierCore(index, lwidth=32, rwidth=32, lsigned=True, rsigned=True, depth=6):
    retwidth = lwidth + rwidth
    
    m = module.Module('multiplier_core_%d' % index)
    
    clk = m.Input('CLK')
    update = m.Input('update')
    
    a = m.Input('a', lwidth)
    b = m.Input('b', rwidth)
    c = m.Output('c', retwidth)
    
    _a = m.Reg('_a', lwidth, signed=lsigned)
    _b = m.Reg('_b', rwidth, signed=rsigned)
    tmpval = [ m.Reg('_tmpval%d' % i, retwidth, signed=True) for i in range(depth-1) ]
    rslt = m.Wire('rslt', retwidth, signed=True)

    __a = _a
    __b = _b
    if not lsigned:
        __a = vtypes.SystemTask('signed', vtypes.Cat(vtypes.Int(0, width=1), _a))
    if not rsigned:
        __b = vtypes.SystemTask('signed', vtypes.Cat(vtypes.Int(0, width=1), _b))
    
    m.Assign( rslt(__a * __b) )
    m.Assign( c(tmpval[depth-2]) )
    
    m.Always(vtypes.Posedge(clk))(
        vtypes.If(update)(
            _a(a),
            _b(b),
            tmpval[0](rslt),
            [ tmpval[i](tmpval[i-1]) for i in range(1, depth-1) ]
        ))
    
    return m

def mkMultiplier(index, lwidth=32, rwidth=32, lsigned=True, rsigned=True, depth=6):
    if lwidth < 0: raise ValueError("data width must be greater than 0.")
    if rwidth < 0: raise ValueError("data width must be greater than 0.")
    if depth < 2: raise ValueError("depth must be greater than 2.")
    
    retwidth = lwidth + rwidth
    
    mult = mkMultiplierCore(index, lwidth, rwidth, lsigned, rsigned, depth)
    
    m = module.Module('multiplier_%d' % index)
    
    clk = m.Input('CLK')
    rst = m.Input('RST')
    
    update = m.Input('update')
    enable = m.Input('enable')
    valid = m.Output('valid')
    
    a = m.Input('a', lwidth)
    b = m.Input('b', rwidth)
    c = m.Output('c', retwidth)

    valid_reg = [ m.Reg('valid_reg%d' % i) for i in range(depth) ]

    m.Assign( valid(valid_reg[depth-1]) )

    m.Always(vtypes.Posedge(clk))(
        vtypes.If(rst)(
            [ valid_reg[i](0) for i in range(depth) ]
        ).Else(
            vtypes.If(update)(
                valid_reg[0](enable),
                [ valid_reg[i](valid_reg[i-1]) for i in range(1, depth) ]
            )
        ))
    
    ports = [ ('CLK', clk), ('update', update), ('a', a), ('b', b), ('c', c) ]
    m.Instance(mult, 'mult', ports=ports)
    
    return m

# global multiplier count
index_count = 0
def get_mul(lwidth=32, rwidth=32, lsigned=True, rsigned=True, depth=6):
    global index_count
    mul = mkMultiplier(index_count, lwidth, rwidth, lsigned, rsigned, depth)
    index_count += 1
    return mul

def reset():
    global index_count
    index_count = 0



from __future__ import absolute_import
from __future__ import print_function
import veriloggen
import pipeline_acc_add

expected_verilog = """
module test;
  reg CLK;
  reg RST;
  reg [32-1:0] x;
  wire [32-1:0] y;
  reg prst;

  blinkled
  uut
  (
    .CLK(CLK),
    .RST(RST),
    .x(x),
    .y(y),
    .prst(prst)
  );

  reg reset_done;

  initial begin
    $dumpfile("uut.vcd");
    $dumpvars(0, uut);
  end


  initial begin
    CLK = 0;
    forever begin
      #5 CLK = (!CLK);
    end
  end


  initial begin
    RST = 0;
    reset_done = 0;
    prst = 0;
    x = 0;
    #100;
    RST = 1;
    #100;
    RST = 0;
    #1000;
    reset_done = 1;
    @(posedge CLK);
    #1;
    #10000;
    $finish;
  end

  reg [32-1:0] _tmp_0;
  reg [32-1:0] xfsm;
  localparam xfsm_init = 0;
  localparam xfsm_1 = 1;
  localparam xfsm_2 = 2;
  localparam xfsm_3 = 3;
  localparam xfsm_4 = 4;
  localparam xfsm_5 = 5;
  localparam xfsm_6 = 6;
  localparam xfsm_7 = 7;

  always @(posedge CLK) begin
    if(RST) begin
      xfsm <= xfsm_init;
      _tmp_0 <= 0;
    end else begin
      case(xfsm)
        xfsm_init: begin
          if(reset_done) begin
            xfsm <= xfsm_1;
          end 
        end
        xfsm_1: begin
          x <= (x + 1);
          _tmp_0 <= (_tmp_0 + 1);
          if((_tmp_0 == 10)) begin
            xfsm <= xfsm_2;
          end 
        end
        xfsm_2: begin
          x <= 0;
          xfsm <= xfsm_3;
        end
        xfsm_3: begin
          xfsm <= xfsm_4;
        end
        xfsm_4: begin
          xfsm <= xfsm_5;
        end
        xfsm_5: begin
          xfsm <= xfsm_6;
        end
        xfsm_6: begin
          xfsm <= xfsm_7;
        end
        xfsm_7: begin
          $finish;
        end
      endcase
    end
  end

  always @(posedge CLK) begin
    if(reset_done) begin
      $display("x=%d", x);
      $display("y=%d", y);
    end 
  end

endmodule


module blinkled
(
  input CLK,
  input RST,
  input [32-1:0] x,
  output [32-1:0] y,
  input prst
);

  reg [32-1:0] _df_data_0;
  assign y = _df_data_0;

  always @(posedge CLK) begin
    if(RST) begin
      _df_data_0 <= 0;
    end else begin
      _df_data_0 <= _df_data_0 + x;
      if(prst) begin
        _df_data_0 <= 0;
      end 
    end
  end

endmodule
"""

def test():
    veriloggen.reset()
    test_module = pipeline_acc_add.mkTest()
    code = test_module.to_verilog()

    from pyverilog.vparser.parser import VerilogParser
    from pyverilog.ast_code_generator.codegen import ASTCodeGenerator
    parser = VerilogParser()
    expected_ast = parser.parse(expected_verilog)
    codegen = ASTCodeGenerator()
    expected_code = codegen.visit(expected_ast)

    assert(expected_code == code)



from __future__ import absolute_import
from __future__ import print_function

import veriloggen.core.vtypes as vtypes
import veriloggen.core.module as module
import veriloggen.types.fifo as fifo
from veriloggen.seq.seq import Seq


def CoramChannelWriteInterface(m, name=None, datawidth=32,
                               p_enq='ENQ', p_wdata='D',
                               p_full='FULL', p_almost_full='ALM_FULL'):
    return fifo.FifoWriteInterface(
        m, name=name, datawidth=datawidth,
        p_enq=p_enq, p_wdata=p_wdata, p_full=p_full, p_almost_full=p_almost_full)


def CoramChannelReadInterface(m, name=None, datawidth=32,
                              p_deq='DEQ', p_rdata='Q',
                              p_empty='EMPTY', p_almost_empty='ALM_EMPTY'):
    return fifo.FifoReadInterface(
        m, name=name, datawidth=datawidth,
        p_deq=p_deq, p_rdata=p_rdata, p_empty=p_empty, p_almost_empty=p_almost_empty)


def CoramChannelWriteMasterInterface(m, name=None, datawidth=32,
                                     p_enq='ENQ', p_wdata='D',
                                     p_full='FULL', p_almost_full='ALM_FULL'):
    return fifo.FifoWriteMasterInterface(
        m, name=name, datawidth=datawidth,
        p_enq=p_enq, p_wdata=p_wdata, p_full=p_full, p_almost_full=p_almost_full)


def CoramChannelReadMasterInterface(m, name=None, datawidth=32,
                                    p_deq='DEQ', p_rdata='Q',
                                    p_empty='EMPTY', p_almost_empty='ALM_EMPTY'):
    return fifo.FifoReadMasterInterface(
        m, name=name, datawidth=datawidth,
        p_deq=p_deq, p_rdata=p_rdata, p_empty=p_empty, p_almost_empty=p_almost_empty)


def CoramChannelWriteSlaveInterface(m, name=None, datawidth=32,
                                    p_enq='ENQ', p_wdata='D',
                                    p_full='FULL', p_almost_full='ALM_FULL'):
    return fifo.FifoWriteSlaveInterface(
        m, name=name, datawidth=datawidth,
        p_enq=p_enq, p_wdata=p_wdata, p_full=p_full, p_almost_full=p_almost_full)


def CoramChannelReadSlaveInterface(m, name=None, datawidth=32,
                                   p_deq='DEQ', p_rdata='Q',
                                   p_empty='EMPTY', p_almost_empty='ALM_EMPTY'):
    return fifo.FifoReadSlaveInterface(
        m, name=name, datawidth=datawidth,
        p_deq=p_deq, p_rdata=p_rdata, p_empty=p_empty, p_almost_empty=p_almost_empty)


def mkCoramChannelDefinition():
    name = 'CoramChannel'
    m = module.Module(name)

    coram_thread_name = m.Parameter('CORAM_THREAD_NAME', 'none')
    coram_thread_id = m.Parameter('CORAM_THREAD_ID', 0)
    coram_id = m.Parameter('CORAM_ID', 0)
    coram_sub_id = m.Parameter('CORAM_SUB_ID', 0)
    coram_addr_len = m.Parameter('CORAM_ADDR_LEN', 10)
    coram_data_width = m.Parameter('CORAM_DATA_WIDTH', 32)

    clk = m.Input('CLK')
    rst = m.Input('RST')

    wif = CoramChannelWriteSlaveInterface(m, datawidth=coram_data_width)
    rif = CoramChannelReadSlaveInterface(m, datawidth=coram_data_width)

    wif.full.assign(0)
    wif.almost_full.assign(0)
    rif.rdata.assign(1)
    rif.empty.assign(0)
    rif.almost_empty.assign(0)

    return m


coram_channel_definition = None


def get_coram_channel_definition():
    global coram_channel_definition
    if coram_channel_definition is None:
        coram_channel_definition = mkCoramChannelDefinition()
    return coram_channel_definition


class CoramChannel(fifo.Fifo):

    def __init__(self, m, clk, rst,
                 thread_name, thread_id, id, sub_id=None,
                 datawidth=32, addrwidth=4):

        self.m = m
        self.clk = clk
        self.rst = rst
        self.thread_name = thread_name
        self.thread_id = thread_id
        self.id = id
        self.sub_id = sub_id
        self.name = '_'.join([thread_name, 'channel', str(id)])
        self.datawidth = datawidth
        self.addrwidth = addrwidth
        self.wif = CoramChannelWriteInterface(self.m, self.name, datawidth)
        self.rif = CoramChannelReadInterface(self.m, self.name, datawidth)
        self.definition = get_coram_channel_definition()

        params = []
        params.append(('CORAM_THREAD_NAME', self.thread_name))
        params.append(('CORAM_THREAD_ID', self.thread_id))
        params.append(('CORAM_ID', self.id))
        if self.sub_id is not None:
            params.append(('CORAM_SUB_ID', self.sub_id))
        params.append(('CORAM_ADDR_LEN', self.addrwidth))
        params.append(('CORAM_DATA_WIDTH', self.datawidth))

        ports = []
        ports.append(('CLK', self.clk))
        ports.append(('RST', self.rst))
        ports.extend(m.connect_ports(self.definition, prefix=self.name + '_'))

        self.inst = self.m.Instance(
            self.definition, 'inst_' + self.name, params, ports)

        self.seq = Seq(m, self.name, clk, rst)

        self._max_size = (2 ** self.addrwidth - 1 if isinstance(self.addrwidth, int) else
                          vtypes.Int(2) ** self.addrwidth - 1)

        self._count = 0

        self._enq_disabled = False
        self._deq_disabled = False
